from __future__ import annotations

import json
import pathlib
from base64 import b64encode
from typing import Any

import pytest
from fontTools.misc.transform import Transform

from ufoLib2.constants import DATA_LIB_KEY
from ufoLib2.objects import (
    Anchor,
    Component,
    Contour,
    DataSet,
    Font,
    Glyph,
    Guideline,
    Image,
    ImageSet,
    Info,
    Kerning,
    Layer,
    LayerSet,
    Lib,
    Point,
)
from ufoLib2.objects.info import (
    GaspBehavior,
    GaspRangeRecord,
    NameRecord,
    WidthClass,
    WoffMetadataCopyright,
    WoffMetadataCredit,
    WoffMetadataCredits,
    WoffMetadataDescription,
    WoffMetadataExtension,
    WoffMetadataExtensionItem,
    WoffMetadataExtensionName,
    WoffMetadataExtensionValue,
    WoffMetadataLicense,
    WoffMetadataLicensee,
    WoffMetadataText,
    WoffMetadataTrademark,
    WoffMetadataUniqueID,
    WoffMetadataVendor,
)

# isort: off
cattrs = pytest.importorskip("cattrs")
from ufoLib2.converters import register_hooks, structure, unstructure  # noqa: E402


@pytest.mark.parametrize(
    "obj, expected",
    [
        (Anchor(0, 0), {"x": 0, "y": 0}),
        (
            Anchor(1, -2, "top", "1,0,0,1", "01234"),
            {"x": 1, "y": -2, "name": "top", "color": "1,0,0,1", "identifier": "01234"},
        ),
        (Point(0, 0), {"x": 0, "y": 0}),
        (
            Point(1.5, -2.1, "qcurve", True, "foobar", "12345"),
            {
                "x": 1.5,
                "y": -2.1,
                "type": "qcurve",
                "smooth": True,
                "name": "foobar",
                "identifier": "12345",
            },
        ),
        (Component("a"), {"baseGlyph": "a"}),
        (
            Component("b", Transform(1, 0, 0, 0.5, 10, 20), "abcd1234"),
            {
                "baseGlyph": "b",
                "transformation": (1, 0, 0, 0.5, 10, 20),
                "identifier": "abcd1234",
            },
        ),
        (Contour(), {}),
        (
            Contour([Point(0, 0), Point(1, 2)], "zzzz"),
            {"points": [{"x": 0, "y": 0}, {"x": 1, "y": 2}], "identifier": "zzzz"},
        ),
        (Image(), {}),
        (
            Image("path/to/image.png", Transform(2, 0, 0, 2, 0, 0), "1,1,1,1"),
            {
                "fileName": "path/to/image.png",
                "transformation": (2, 0, 0, 2, 0, 0),
                "color": "1,1,1,1",
            },
        ),
        (Lib(), {}),
        (Lib(foo="bar"), {"foo": "bar"}),
        (Lib(foo=[1, 2], bar={"baz": 3}), {"foo": [1, 2], "bar": {"baz": 3}}),
        (
            Lib(foo={"bar": b"baz"}, oof=[b"rab", "zab"]),
            {
                "foo": {
                    "bar": {"data": "YmF6", "type": DATA_LIB_KEY},
                },
                "oof": [
                    {"data": "cmFi", "type": DATA_LIB_KEY},
                    "zab",
                ],
            },
        ),
        (Guideline(x=0, name="foo"), {"x": 0, "name": "foo"}),
        (Guideline(y=1, name="bar"), {"y": 1, "name": "bar"}),
        (
            Guideline(
                x=1, y=2, angle=45.0, name="baz", color="0,1,0.5,1", identifier="0001"
            ),
            {
                "x": 1,
                "y": 2,
                "angle": 45.0,
                "name": "baz",
                "color": "0,1,0.5,1",
                "identifier": "0001",
            },
        ),
        (Glyph(), {}),
        (
            Glyph(
                "a",
                width=1000,
                height=800,
                unicodes=[97],
                lib=Lib(foo=b"bar"),
                note="Latin lowercase 'a'",
                anchors=[Anchor(100, 200, "top"), Anchor(100, -50, "bottom")],
                contours=[
                    Contour(
                        [Point(0, 0, "line"), Point(1, 1, "line"), Point(1, 0, "line")]
                    ),
                    Contour(
                        [
                            Point(2, 2, "move"),
                            Point(3, 2),
                            Point(4, 1),
                            Point(4, -1, "curve"),
                        ]
                    ),
                ],
                guidelines=[Guideline(x=10)],
            ),
            {
                "name": "a",
                "width": 1000,
                "height": 800,
                "unicodes": [97],
                "lib": {"foo": {"data": "YmFy", "type": DATA_LIB_KEY}},
                "note": "Latin lowercase 'a'",
                "anchors": [
                    {"name": "top", "x": 100, "y": 200},
                    {"name": "bottom", "x": 100, "y": -50},
                ],
                "contours": [
                    {
                        "points": [
                            {"x": 0, "y": 0, "type": "line"},
                            {"x": 1, "y": 1, "type": "line"},
                            {"x": 1, "y": 0, "type": "line"},
                        ]
                    },
                    {
                        "points": [
                            {"x": 2, "y": 2, "type": "move"},
                            {"x": 3, "y": 2},
                            {"x": 4, "y": 1},
                            {"x": 4, "y": -1, "type": "curve"},
                        ]
                    },
                ],
                "guidelines": [{"x": 10}],
            },
        ),
        (Kerning(), {}),
        (
            Kerning({("a", "b"): -10, ("a", "d"): 5, ("b", "d"): 0}),
            {"a": {"b": -10, "d": 5}, "b": {"d": 0}},
        ),
        (
            Info(
                familyName="Test",
                styleName="Bold",
                versionMajor=2,
                versionMinor=1,
                unitsPerEm=1000,
                guidelines=[Guideline(y=500, name="x-height")],
                openTypeGaspRangeRecords=[
                    GaspRangeRecord(
                        rangeMaxPPEM=18,
                        rangeGaspBehavior=[GaspBehavior.SYMMETRIC_SMOOTHING],
                    )
                ],
                openTypeHeadFlags=[0, 1, 2],
                openTypeNameRecords=[
                    NameRecord(
                        nameID=5,
                        platformID=0,
                        encodingID=0,
                        languageID=0,
                        string="Version 2.1",
                    )
                ],
                openTypeOS2WidthClass=WidthClass.MEDIUM,
                postscriptBlueScale=0.25,
                woffMajorVersion=1,
                woffMinorVersion=0,
                woffMetadataUniqueID=WoffMetadataUniqueID(
                    "com.example.fontvendor.demofont.rev12345"
                ),
                woffMetadataVendor=WoffMetadataVendor(
                    name="Font Vendor",
                    url="http://fontvendor.example.com",
                ),
                woffMetadataCredits=WoffMetadataCredits(
                    [
                        WoffMetadataCredit(
                            name="Font Designer",
                            url="http://fontdesigner.example.com",
                            role="Lead",
                        ),
                        WoffMetadataCredit(
                            name="Another Font Designer",
                            url="http://anotherdesigner.example.com",
                            role="Contributor",
                        ),
                    ]
                ),
                woffMetadataDescription=WoffMetadataDescription(
                    text=[
                        WoffMetadataText(
                            "A member of the Demo font family...", language="en"
                        )
                    ]
                ),
                woffMetadataLicense=WoffMetadataLicense(
                    id="fontvendor-Web-corporate-v2",
                    url="http://fontvendor.example.com/license",
                    text=[
                        WoffMetadataText("A license goes here", language="en"),
                        WoffMetadataText("Un permis va ici", language="fr"),
                    ],
                ),
                woffMetadataCopyright=WoffMetadataCopyright(
                    [
                        WoffMetadataText("Copyright ©2009 Font Vendor", language="en"),
                        WoffMetadataText("저작권 ©2009 Font Vendor", language="ko"),
                    ],
                ),
                woffMetadataTrademark=WoffMetadataTrademark(
                    [
                        WoffMetadataText(
                            "Demo Font is a trademark of Font Vendor", language="en"
                        ),
                        WoffMetadataText(
                            "Demo Font est une marque déposée de Font Vendor",
                            language="fr",
                        ),
                    ]
                ),
                woffMetadataLicensee=WoffMetadataLicensee(
                    "Wonderful Websites, Inc.",
                ),
                woffMetadataExtensions=[
                    WoffMetadataExtension(
                        id="org.example.fonts.metadata.v1",
                        names=[
                            WoffMetadataExtensionName(
                                "Additional font information", language="en"
                            ),
                            WoffMetadataExtensionName(
                                "L'information supplémentaire de fonte", language="fr"
                            ),
                        ],
                        items_=[
                            WoffMetadataExtensionItem(
                                id="org.example.fonts.metadata.v1.why",
                                names=[
                                    WoffMetadataExtensionName("Purpose", language="en"),
                                    WoffMetadataExtensionName("But", language="fr"),
                                ],
                                values_=[
                                    WoffMetadataExtensionValue(
                                        "An example of WOFF packaging", language="en"
                                    ),
                                    WoffMetadataExtensionValue(
                                        "Un exemple de l'empaquetage de WOFF",
                                        language="fr",
                                    ),
                                ],
                            )
                        ],
                    )
                ],
            ),
            {
                "familyName": "Test",
                "styleName": "Bold",
                "versionMajor": 2,
                "versionMinor": 1,
                "unitsPerEm": 1000,
                "guidelines": [{"y": 500, "name": "x-height"}],
                "openTypeGaspRangeRecords": [
                    {"rangeMaxPPEM": 18, "rangeGaspBehavior": [3]},
                ],
                "openTypeHeadFlags": [0, 1, 2],
                "openTypeNameRecords": [
                    {
                        "nameID": 5,
                        "platformID": 0,
                        "encodingID": 0,
                        "languageID": 0,
                        "string": "Version 2.1",
                    }
                ],
                "openTypeOS2WidthClass": 5,
                "postscriptBlueScale": 0.25,
                "woffMajorVersion": 1,
                "woffMinorVersion": 0,
                "woffMetadataUniqueID": {
                    "id": "com.example.fontvendor.demofont.rev12345"
                },
                "woffMetadataVendor": {
                    "name": "Font Vendor",
                    "url": "http://fontvendor.example.com",
                },
                "woffMetadataCredits": {
                    "credits": [
                        {
                            "name": "Font Designer",
                            "role": "Lead",
                            "url": "http://fontdesigner.example.com",
                        },
                        {
                            "name": "Another Font Designer",
                            "role": "Contributor",
                            "url": "http://anotherdesigner.example.com",
                        },
                    ]
                },
                "woffMetadataDescription": {
                    "text": [
                        {
                            "language": "en",
                            "text": "A member of the Demo font family...",
                        }
                    ]
                },
                "woffMetadataLicense": {
                    "id": "fontvendor-Web-corporate-v2",
                    "text": [
                        {"language": "en", "text": "A license goes here"},
                        {"language": "fr", "text": "Un permis va ici"},
                    ],
                    "url": "http://fontvendor.example.com/license",
                },
                "woffMetadataCopyright": {
                    "text": [
                        {"language": "en", "text": "Copyright ©2009 Font Vendor"},
                        {"language": "ko", "text": "저작권 ©2009 Font Vendor"},
                    ]
                },
                "woffMetadataTrademark": {
                    "text": [
                        {
                            "language": "en",
                            "text": "Demo Font is a trademark of Font Vendor",
                        },
                        {
                            "language": "fr",
                            "text": "Demo Font est une marque déposée "
                            "de Font Vendor",
                        },
                    ]
                },
                "woffMetadataLicensee": {"name": "Wonderful Websites, Inc."},
                "woffMetadataExtensions": [
                    {
                        "id": "org.example.fonts.metadata.v1",
                        "names": [
                            {"language": "en", "text": "Additional font information"},
                            {
                                "language": "fr",
                                "text": "L'information supplémentaire de fonte",
                            },
                        ],
                        "items": [
                            {
                                "id": "org.example.fonts.metadata.v1.why",
                                "names": [
                                    {"language": "en", "text": "Purpose"},
                                    {"language": "fr", "text": "But"},
                                ],
                                "values": [
                                    {
                                        "language": "en",
                                        "text": "An example of WOFF packaging",
                                    },
                                    {
                                        "language": "fr",
                                        "text": "Un exemple de l'empaquetage de WOFF",
                                    },
                                ],
                            }
                        ],
                    }
                ],
            },
        ),
        # 'public.default' is a special case, default=True by definition
        (Layer(), {"name": "public.default"}),
        (Layer("foo", default=True), {"name": "foo", "default": True}),
        (Layer("bar"), {"name": "bar"}),
        (
            Layer(
                name="foreground",
                glyphs={"a": Glyph(), "b": Glyph()},
                color="1,0,1,1",
                lib=Lib(foobar=0.1),
                default=True,
            ),
            {
                "name": "foreground",
                "glyphs": {"a": {}, "b": {}},
                "color": "1,0,1,1",
                "lib": {"foobar": 0.1},
                "default": True,
            },
        ),
        (LayerSet.default(), [{"name": "public.default"}]),
        (
            LayerSet.from_iterable(
                [Layer("foreground"), Layer("background")],
                defaultLayerName="foreground",  # deprecated
            ),
            [{"name": "foreground", "default": True}, {"name": "background"}],
        ),
        (
            LayerSet.from_iterable(
                [Layer("foreground", default=True), Layer("background")],
            ),
            [{"name": "foreground", "default": True}, {"name": "background"}],
        ),
        (DataSet(), {}),
        (DataSet({"foo": b"bar"}), {"foo": "YmFy"}),
        (ImageSet(), {}),
        (ImageSet({"foo": b"bar"}), {"foo": "YmFy"}),
        (Font(), {"layers": [{"name": "public.default"}]}),
        (
            Font(
                layers=[
                    Layer(name="foreground", default=True),
                    Layer(name="background"),
                ]
            ),
            {
                "layers": [
                    {"name": "foreground", "default": True},
                    {"name": "background"},
                ],
            },
        ),
        (
            Font(
                layers=[
                    Layer(
                        glyphs=[Glyph("a", tempLib={"foo": "bar"})],
                        tempLib={"a": 123},
                    )
                ],
                info=Info(familyName="Test"),
                features="languagesystem DFLT dflt;",
                groups={"LOWERCASE": ["a"]},
                kerning={("a", "a"): 10},
                lib={"foo": "bar"},
                data={"baz": b"\0"},
                images={"foobarbaz": b"\0"},
                tempLib={"foofoo": "barbar"},
            ),
            {
                "layers": [
                    {
                        "name": "public.default",
                        "glyphs": {"a": {"tempLib": {"foo": "bar"}}},
                        "tempLib": {"a": 123},
                    }
                ],
                "info": {"familyName": "Test"},
                "features": "languagesystem DFLT dflt;",
                "groups": {"LOWERCASE": ["a"]},
                "kerning": {"a": {"a": 10}},
                "lib": {"foo": "bar"},
                "data": {"baz": "AA=="},
                "images": {"foobarbaz": "AA=="},
                "tempLib": {"foofoo": "barbar"},
            },
        ),
    ],
)
def test_unstructure_structure(obj: Any, expected: dict[str, Any]) -> None:
    assert unstructure(obj) == expected
    assert structure(expected, type(obj)) == obj


def test_unstructure_lazy_font(ufo_UbuTestData: Font) -> None:
    font1 = ufo_UbuTestData
    assert font1._lazy

    font_data = unstructure(font1)

    font2 = structure(font_data, Font)
    assert font2 == font1

    assert not font2._lazy


@pytest.mark.parametrize("forbid_extra_keys", [True, False])
def test_structure_forbid_extra_keys(forbid_extra_keys: bool) -> None:
    conv = cattrs.Converter(
        forbid_extra_keys=forbid_extra_keys,
        detailed_validation=False,
    )
    register_hooks(conv)
    data = {"name": "a", "foo": "bar"}
    if forbid_extra_keys:
        with pytest.raises(
            cattrs.errors.ForbiddenExtraKeysError,
            match="Extra fields in constructor for .*: foo",
        ):
            conv.structure(data, Glyph)
    else:
        assert conv.structure(data, Glyph) == Glyph(name="a")


@pytest.mark.parametrize(
    "omit_if_default, obj, expected",
    [
        pytest.param(
            True, Anchor(x=1.0, y=2.0), {"x": 1.0, "y": 2.0}, id="True-Anchor"
        ),
        pytest.param(
            False, Anchor(x=1.0, y=2.0), {"x": 1.0, "y": 2.0}, id="False-Anchor"
        ),
        pytest.param(True, Component("foo"), {"baseGlyph": "foo"}, id="True-Component"),
        pytest.param(
            False,
            Component("foo"),
            {"baseGlyph": "foo", "transformation": (1, 0, 0, 1, 0, 0)},
            id="False-Component",
        ),
        pytest.param(True, Contour(), {}, id="True-Contour"),
        pytest.param(False, Contour(), {"points": []}, id="False-Contour"),
        pytest.param(True, Point(x=1.0, y=2.0), {"x": 1.0, "y": 2.0}, id="True-Point"),
        pytest.param(
            False,
            Point(x=1.0, y=2.0),
            {"x": 1.0, "y": 2.0, "smooth": False},
            id="False-Point",
        ),
        pytest.param(True, Glyph(), {}, id="True-Glyph"),
        pytest.param(
            False,
            Glyph(),
            {
                "width": 0,
                "height": 0,
                "unicodes": [],
                "image": {"transformation": (1, 0, 0, 1, 0, 0)},
                "lib": {},
                "anchors": [],
                "components": [],
                "contours": [],
                "guidelines": [],
                "tempLib": {},
            },
            id="False-Glyph",
        ),
        pytest.param(True, Layer(), {"name": "public.default"}, id="True-Layer"),
        pytest.param(
            False,
            Layer(),
            {
                "name": "public.default",
                "default": True,
                "glyphs": {},
                "lib": {},
                "tempLib": {},
            },
            id="False-Layer",
        ),
        pytest.param(
            True, Font(), {"layers": [{"name": "public.default"}]}, id="True-Font"
        ),
        pytest.param(
            False,
            Font(),
            {
                "data": {},
                "features": "",
                "groups": {},
                "images": {},
                "info": {},
                "kerning": {},
                "layers": [
                    {
                        "default": True,
                        "glyphs": {},
                        "lib": {},
                        "name": "public.default",
                        "tempLib": {},
                    }
                ],
                "lib": {},
                "tempLib": {},
            },
            id="False-Font",
        ),
    ],
)
def test_omit_if_default(obj: Any, expected: Any, omit_if_default: bool) -> None:
    conv = cattrs.Converter(omit_if_default=omit_if_default)
    register_hooks(conv)
    assert conv.unstructure(obj) == expected


@pytest.mark.parametrize(
    "allow_bytes, obj, expected",
    [
        pytest.param(True, b"foo", b"foo", id="True-bytes"),
        pytest.param(False, b"foo", b64encode(b"foo").decode(), id="False-bytes"),
        pytest.param(
            True, DataSet({"foo": b"bar"}), {"foo": b"bar"}, id="True-DataSet"
        ),
        pytest.param(
            False,
            DataSet({"foo": b"bar"}),
            {"foo": b64encode(b"bar").decode()},
            id="False-DataSet",
        ),
        pytest.param(
            True,
            ImageSet({"foo.png": b"\x89PNG"}),
            {"foo.png": b"\x89PNG"},
            id="True-ImageSet",
        ),
        pytest.param(
            False,
            ImageSet({"foo.png": b"\x89PNG"}),
            {"foo.png": b64encode(b"\x89PNG").decode()},
            id="False-ImageSet",
        ),
        pytest.param(
            True,
            Lib(foo={"bar": b"baz"}, oof=[b"rab", "zab"]),
            {"foo": {"bar": b"baz"}, "oof": [b"rab", "zab"]},
            id="True-Lib",
        ),
        pytest.param(
            False,
            Lib(foo={"bar": b"baz"}, oof=[b"rab", "zab"]),
            {
                "foo": {
                    "bar": {"data": b64encode(b"baz").decode(), "type": DATA_LIB_KEY},
                },
                "oof": [
                    {"data": b64encode(b"rab").decode(), "type": DATA_LIB_KEY},
                    "zab",
                ],
            },
            id="False-Lib",
        ),
    ],
)
def test_allow_bytes(obj: Any, expected: Any, allow_bytes: bool) -> None:
    conv = cattrs.Converter()
    register_hooks(conv, allow_bytes=allow_bytes)

    assert conv.unstructure(obj) == expected
    assert conv.structure(expected, type(obj)) == obj


def test_custom_type_overrides() -> None:
    conv = cattrs.Converter(type_overrides={Image: cattrs.override(omit=True)})
    register_hooks(conv)

    # check that Glyph.image attribute (of type Image) is omitted
    assert conv.unstructure(Glyph()) == {
        "width": 0,
        "height": 0,
        "unicodes": [],
        "lib": {},
        "anchors": [],
        "components": [],
        "contours": [],
        "guidelines": [],
        "tempLib": {},
    }


def test_json_dumps(datadir: pathlib.Path) -> None:
    font = Font.open(datadir / "MutatorSansBoldCondensed.ufo")
    # need to get rid of CR/LF newlines that sneak in the features.fea when
    # opening the UFO on Windows
    font.features.normalize_newlines()

    data = unstructure(font)

    expected = (datadir / "MutatorSansBoldCondensed.json").read_text()

    assert json.dumps(data, indent=2, sort_keys=True) == expected


def test_json_loads(datadir: pathlib.Path) -> None:
    data = json.loads((datadir / "MutatorSansBoldCondensed.json").read_bytes())

    expected = Font.open(datadir / "MutatorSansBoldCondensed.ufo")
    # need to get rid of CR/LF newlines that sneak in the features.fea when
    # opening the UFO on Windows
    expected.features.normalize_newlines()

    assert structure(data, Font) == expected



from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from typing import Any, Type

import pytest
from fontTools import ufoLib

import ufoLib2
import ufoLib2.objects
from ufoLib2.objects import Features, Font, Glyph, Kerning, Layer, LayerSet, Lib
from ufoLib2.objects.layerSet import _LAYER_NOT_LOADED
from ufoLib2.objects.misc import _DATA_NOT_LOADED


def test_import_version() -> None:
    assert hasattr(ufoLib2, "__version__")
    assert isinstance(ufoLib2.__version__, str)


def test_LayerSet_load_layers_on_iteration(tmp_path: Path) -> None:
    ufo = ufoLib2.Font()
    ufo.layers.newLayer("test")
    ufo_save_path = tmp_path / "test.ufo"
    ufo.save(ufo_save_path)
    ufo = ufoLib2.Font.open(ufo_save_path)
    assert set(ufo.layers.keys()) == {"public.default", "test"}
    for layer in ufo.layers:
        assert layer is not _LAYER_NOT_LOADED


def test_lazy_data_loading_saveas(ufo_UbuTestData: Font, tmp_path: Path) -> None:
    ufo = ufo_UbuTestData
    ufo_path = tmp_path / "UbuTestData2.ufo"
    ufo.save(ufo_path)
    assert all(v is not _DATA_NOT_LOADED for v in ufo.data._data.values())


def test_lazy_data_loading_inplace_no_load(ufo_UbuTestData: Font) -> None:
    ufo = ufo_UbuTestData
    ufo.save()
    assert all(v is _DATA_NOT_LOADED for v in ufo.data._data.values())


def test_lazy_data_loading_inplace_load_some(ufo_UbuTestData: Font) -> None:
    ufo = ufo_UbuTestData
    some_data = b"abc"
    ufo.data["com.github.fonttools.ttx/T_S_I__0.ttx"] = some_data
    ufo.save()
    assert all(
        v is _DATA_NOT_LOADED for k, v in ufo.data._data.items() if "T_S_I__0" not in k
    )
    assert ufo.data["com.github.fonttools.ttx/T_S_I__0.ttx"] == some_data


def test_deepcopy_lazy_object(datadir: Path) -> None:
    path = datadir / "UbuTestData.ufo"
    font1 = ufoLib2.Font.open(path, lazy=True)

    font2 = deepcopy(font1)

    assert font1 is not font2
    assert font1 == font2

    assert font1.layers is not font2.layers
    assert font1.layers == font2.layers

    assert font1.layers.defaultLayer is not font2.layers.defaultLayer
    assert font1.layers.defaultLayer == font2.layers.defaultLayer

    assert font1.data is not font2.data
    assert font1.data == font2.data

    assert font1.images is not font2.images
    assert font1.images == font2.images

    assert font1.reader is not None
    assert not font1.reader.fs.isclosed()
    assert not font1._lazy

    assert font2.reader is None
    assert not font2._lazy

    assert font1.path == path
    assert font2.path is None


def test_unlazify(datadir: Path) -> None:
    reader = ufoLib.UFOReader(datadir / "UbuTestData.ufo")
    font = ufoLib2.Font.read(reader, lazy=True)

    assert font._reader is reader
    assert not reader.fs.isclosed()

    font.unlazify()

    assert font._lazy is False


def test_auto_unlazify_font(datadir: Path) -> None:
    font1 = ufoLib2.Font.open(datadir / "UbuTestData.ufo", lazy=True)
    font2 = ufoLib2.Font.open(datadir / "UbuTestData.ufo", lazy=False)

    assert font1 == font2


def test_auto_unlazify_data(datadir: Path) -> None:
    font1 = ufoLib2.Font.open(datadir / "UbuTestData.ufo", lazy=True)
    font2 = ufoLib2.Font.open(datadir / "UbuTestData.ufo", lazy=False)

    assert font1.data == font2.data


def test_auto_unlazify_images(datadir: Path) -> None:
    font1 = ufoLib2.Font.open(datadir / "UbuTestData.ufo", lazy=True)
    font2 = ufoLib2.Font.open(datadir / "UbuTestData.ufo", lazy=False)

    assert font1.images == font2.images


def test_font_eq_and_ne(ufo_UbuTestData: Font) -> None:
    font1 = ufo_UbuTestData
    font2 = deepcopy(font1)

    assert font1 == font2

    font1["a"].contours[0].points[0].x = 0

    assert font1 != font2


def test_empty_layerset() -> None:
    with pytest.raises(ValueError):
        LayerSet(layers={}, defaultLayer=None)  # type: ignore


def test_default_layerset() -> None:
    layers = LayerSet.default()
    assert len(layers) == 1
    assert "public.default" in layers
    assert len(layers["public.default"]) == 0
    assert layers["public.default"].default


def test_custom_layerset() -> None:
    default = Layer()
    ls1 = LayerSet.from_iterable([default])
    assert next(iter(ls1)) is ls1.defaultLayer

    with pytest.raises(ValueError, match="no layer marked as default"):
        LayerSet.from_iterable([Layer(name="abc")])

    with pytest.raises(ValueError, match="no layer marked as default"):
        LayerSet({"abc": Layer(name="abc")})

    with pytest.raises(ValueError, match="more than one layer marked as default"):
        LayerSet.from_iterable([Layer("a", default=True), Layer("b", default=True)])

    with pytest.raises(ValueError, match="more than one layer marked as default"):
        LayerSet({"a": Layer("a", default=True), "b": Layer("b", default=True)})

    with pytest.raises(KeyError, match="duplicate layer name"):
        LayerSet.from_iterable([Layer("a", default=True), Layer("a")])

    with pytest.raises(ValueError, match="default layer .* must be in layer set"):
        LayerSet(layers={"public.default": Layer()}, defaultLayer=Layer())

    # defaultLayerName is deprecated but still works for now
    ls2 = LayerSet.from_iterable([Layer(name="abc")], defaultLayerName="abc")
    assert ls2["abc"].default
    assert ls2["abc"] is ls2.defaultLayer

    # defaultLayer is set automatically based on Layer.default attribute
    ls3 = LayerSet.from_iterable([Layer(name="abc", default=True), Layer("def")])
    assert ls3["abc"].default
    assert ls3["abc"] is ls3.defaultLayer
    assert not ls3["def"].default

    # also for the default constructor, defaultLayer is guessed from Layer.default
    ls4 = LayerSet(layers={"foreground": Layer("foreground", default=True)})
    assert ls4["foreground"] is ls4.defaultLayer

    # unless defaultLayer parameter is set explicitly
    defaultLayer = Layer("foreground", default=True)
    ls5 = LayerSet(layers={"foreground": defaultLayer}, defaultLayer=defaultLayer)
    assert ls5["foreground"] is ls5.defaultLayer
    assert ls5.defaultLayer is defaultLayer


def test_change_default_layer() -> None:
    font = Font(layers=[Layer("foo", default=True), Layer("bar")])
    assert font.layers.defaultLayer is font.layers["foo"]
    assert font.layers["foo"].default
    assert not font.layers["bar"].default

    font.layers.defaultLayer = font.layers["bar"]
    assert font.layers.defaultLayer is font.layers["bar"]
    assert not font.layers["foo"].default
    assert font.layers["bar"].default  # type: ignore

    font.newLayer("baz", default=True)
    assert font.layers.defaultLayer is font.layers["baz"]
    assert font.layers["baz"].default
    assert not font.layers["foo"].default
    assert not font.layers["bar"].default

    font.renameLayer("foo", "public.default")
    assert "foo" not in font.layers
    assert font.layers.defaultLayer is font.layers["public.default"]
    assert font.layers["public.default"].default
    assert not font.layers["baz"].default


def test_change_default_layer_invalid() -> None:
    font = Font(layers=[Layer(), Layer("background")])

    with pytest.raises(
        ValueError,
        match="there's already a layer named 'public.default' which must stay default",
    ):
        font.layers.defaultLayer = font.layers["background"]

    with pytest.raises(
        ValueError, match="Layer .* not found in layer set; can't set as default"
    ):
        font.layers.defaultLayer = Layer("foobar")


def test_guidelines() -> None:
    font = ufoLib2.Font()

    # accept either a mapping or a Guideline object
    font.appendGuideline({"x": 100, "y": 50, "angle": 315})
    font.appendGuideline(ufoLib2.objects.Guideline(x=30))

    assert len(font.guidelines) == 2
    assert font.guidelines == [
        ufoLib2.objects.Guideline(x=100, y=50, angle=315),
        ufoLib2.objects.Guideline(x=30),
    ]

    # setter should clear existing guidelines
    font.guidelines = [{"x": 100}, ufoLib2.objects.Guideline(y=20)]  # type: ignore

    assert len(font.guidelines) == 2
    assert font.guidelines == [
        ufoLib2.objects.Guideline(x=100),
        ufoLib2.objects.Guideline(y=20),
    ]


def test_woff_metadata(datadir: Path, tmp_path: Path) -> None:
    # The WoffMetadataTest.ufo contains all the WOFF metadata, here we check
    # that ufoLib validators accept the data and can read/write fontinfo.plist.
    input_path = datadir / "WoffMetadataTest.ufo"
    output_path = tmp_path / "WoffMetadataTest.ufo"

    font = Font.open(input_path, validate=True)
    font.save(output_path, validate=True)

    assert (input_path / "fontinfo.plist").read_text("utf-8") == (
        (output_path / "fontinfo.plist").read_text("utf-8")
    )


def test_features_normalize_newlines() -> None:
    assert Features("a\r\nb\rc\n").normalize_newlines().text == "a\nb\nc\n"


@pytest.mark.parametrize(
    "klass, attr_name, attr_type, obj",
    [
        (Font, "lib", Lib, {"foo": 1}),
        (Font, "kerning", Kerning, {("a", "b"): -10}),
        (Glyph, "lib", Lib, {"bar": [2, 3]}),
    ],
)
def test_convert_on_setattr(
    klass: Type[Any], attr_name: str, attr_type: Type[Any], obj: Any
) -> None:
    o = klass()
    assert isinstance(getattr(o, attr_name), attr_type)
    assert not isinstance(obj, attr_type)
    setattr(o, attr_name, obj)
    assert isinstance(getattr(o, attr_name), attr_type)



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

import pytest

import ufoLib2.objects

# isort: off
pytest.importorskip("cattrs")

import ufoLib2.serde.json  # noqa: E402


@pytest.mark.parametrize("have_orjson", [False, True], ids=["no-orjson", "with-orjson"])
def test_dumps_loads(
    monkeypatch: Any, have_orjson: bool, ufo_UbuTestData: ufoLib2.objects.Font
) -> None:
    if not have_orjson:
        monkeypatch.setattr(ufoLib2.serde.json, "have_orjson", have_orjson)
    else:
        pytest.importorskip("orjson")

    font = ufo_UbuTestData
    data = font.json_dumps()  # type: ignore

    assert isinstance(data, bytes)

    if have_orjson:
        # with default indent=0, orjson adds no space between keys and values
        assert data[:21] == b'{"layers":[{"name":"p'
    else:
        # built-in json always adds space between keys and values
        assert data[:21] == b'{"layers": [{"name": '

    font2 = ufoLib2.objects.Font.json_loads(data)  # type: ignore

    assert font == font2


@pytest.mark.parametrize("have_orjson", [False, True], ids=["no-orjson", "with-orjson"])
@pytest.mark.parametrize("indent", [None, 2], ids=["no-indent", "indent-2"])
@pytest.mark.parametrize("sort_keys", [False, True], ids=["no-sort-keys", "sort-keys"])
def test_dump_load(
    monkeypatch: Any,
    tmp_path: Path,
    ufo_UbuTestData: ufoLib2.objects.Font,
    have_orjson: bool,
    indent: int | None,
    sort_keys: bool,
) -> None:
    if not have_orjson:
        monkeypatch.setattr(ufoLib2.serde.json, "have_orjson", have_orjson)

    font = ufo_UbuTestData
    with open(tmp_path / "test.json", "wb") as f:
        font.json_dump(f, indent=indent, sort_keys=sort_keys)  # type: ignore

    with open(tmp_path / "test.json", "rb") as f:
        font2 = ufoLib2.objects.Font.json_load(f)  # type: ignore

    assert font == font2

    with open(tmp_path / "test.json", "wb") as f:
        font.json_dump(f, indent=indent, sort_keys=sort_keys)  # type: ignore

    # load/dump work with paths too, not just file objects
    font3 = ufoLib2.objects.Font.json_load(tmp_path / "test.json")  # type: ignore

    assert font == font3

    font.json_dump(  # type: ignore
        tmp_path / "test2.json",
        indent=indent,
        sort_keys=sort_keys,
    )

    assert (tmp_path / "test.json").read_bytes() == (
        tmp_path / "test2.json"
    ).read_bytes()


@pytest.mark.parametrize("indent", [1, 3], ids=["indent-1", "indent-3"])
def test_indent_not_2_orjson(indent: int) -> None:
    pytest.importorskip("orjson")
    with pytest.raises(ValueError):
        ufoLib2.serde.json.dumps(None, indent=indent)


def test_not_allow_bytes(ufo_UbuTestData: ufoLib2.objects.Font) -> None:
    font = ufo_UbuTestData

    # DataSet values are binary data (bytes)
    assert all(isinstance(v, bytes) for v in font.data.values())

    # bytes are are not allowed in JSON, so our converter automatically
    # translates them to Base64 strings upon serializig
    s = font.data.json_dumps()  # type: ignore

    # check that (before structuring the DataSet object) the json data
    # contains in fact str, not bytes
    raw_data = json.loads(s)
    assert isinstance(raw_data, dict)
    assert all(isinstance(v, str) for v in raw_data.values())

    # check that (after structuring the DataSet object) the json data
    # now contains bytes, like the original data
    data = font.data.json_loads(s)  # type: ignore
    assert isinstance(data, ufoLib2.objects.DataSet)
    assert all(isinstance(v, bytes) for v in data.values())



from pathlib import Path

import pytest

import ufoLib2.objects

# isort: off
pytest.importorskip("cattrs")
pytest.importorskip("msgpack")

import msgpack  # type: ignore  # noqa

import ufoLib2.serde.msgpack  # noqa: E402


def test_dumps_loads(ufo_UbuTestData: ufoLib2.objects.Font) -> None:
    font = ufo_UbuTestData
    data = font.msgpack_dumps()  # type: ignore

    assert data[:10] == b"\x85\xa6layers\x92\x82"

    font2 = ufoLib2.objects.Font.msgpack_loads(data)  # type: ignore

    assert font == font2


def test_dump_load(tmp_path: Path, ufo_UbuTestData: ufoLib2.objects.Font) -> None:
    font = ufo_UbuTestData
    with open(tmp_path / "test.msgpack", "wb") as f:
        font.msgpack_dump(f)  # type: ignore

    with open(tmp_path / "test.msgpack", "rb") as f:
        font2 = ufoLib2.objects.Font.msgpack_load(f)  # type: ignore

    assert font == font2

    # laod/dump work with paths too, not just file objects
    font3 = ufoLib2.objects.Font.msgpack_load(tmp_path / "test.msgpack")  # type: ignore

    assert font == font3

    font.msgpack_dump(tmp_path / "test2.msgpack")  # type: ignore

    assert (tmp_path / "test.msgpack").read_bytes() == (
        tmp_path / "test2.msgpack"
    ).read_bytes()


def test_allow_bytes(ufo_UbuTestData: ufoLib2.objects.Font) -> None:
    font = ufo_UbuTestData

    # DataSet values are binary data (bytes)
    assert all(isinstance(v, bytes) for v in font.data.values())

    # bytes *are* allowed in MessagePack (unlike JSON), so its converter should
    # keep them as such (not translate them to Base64 str) upon serializig
    b = font.data.msgpack_dumps()  # type: ignore

    # check that (even before structuring the DataSet object) the msgpack raw data
    # contains in fact bytes, not str
    raw_data = msgpack.unpackb(b)
    assert isinstance(raw_data, dict)
    assert all(isinstance(v, bytes) for v in raw_data.values())

    # of course, also after structuring, the DataSet should contain bytes
    data = font.data.msgpack_loads(b)  # type: ignore
    assert isinstance(data, ufoLib2.objects.DataSet)
    assert all(isinstance(v, bytes) for v in data.values())



import importlib
from typing import Any, Dict, List

import pytest
from attrs import define

import ufoLib2.objects
from ufoLib2.errors import ExtrasNotInstalledError
from ufoLib2.serde import _SERDE_FORMATS_, serde

cattrs = None
try:
    import cattrs  # type: ignore
except ImportError:
    pass


msgpack = None
try:
    import msgpack  # type: ignore
except ImportError:
    pass


EXTRAS_REQUIREMENTS = {
    "json": ["cattrs"],
    "msgpack": ["cattrs", "msgpack"],
}


def assert_extras_not_installed(extras: str, missing_dependency: str) -> None:
    # sanity check that the dependency is not installed
    with pytest.raises(ImportError, match=missing_dependency):
        importlib.import_module(missing_dependency)

    @serde
    @define
    class Foo:
        a: int
        b: str = "bar"

    foo = Foo(1)

    with pytest.raises(
        ExtrasNotInstalledError, match=rf"Extras not installed: ufoLib2\[{extras}\]"
    ) as exc_info:
        dumps_method = getattr(foo, f"{extras}_dumps")
        dumps_method()

    assert isinstance(exc_info.value.__cause__, ModuleNotFoundError)


@pytest.mark.skipif(cattrs is not None, reason="cattrs installed, not applicable")
def test_json_cattrs_not_installed() -> None:
    assert_extras_not_installed("json", "cattrs")


@pytest.mark.skipif(cattrs is not None, reason="cattrs installed, not applicable")
def test_msgpack_cattrs_not_installed() -> None:
    assert_extras_not_installed("msgpack", "cattrs")


@pytest.mark.skipif(msgpack is not None, reason="msgpack installed, not applicable")
def test_msgpack_not_installed() -> None:
    assert_extras_not_installed("msgpack", "msgpack")


BASIC_EMPTY_OBJECTS: List[Dict[str, Any]] = [
    {"class_name": "Anchor", "args": (0, 0)},
    {"class_name": "Component", "args": ("a",)},
    {"class_name": "Contour", "args": ()},
    {"class_name": "DataSet", "args": ()},
    {"class_name": "Features", "args": ()},
    {"class_name": "Font", "args": ()},
    {"class_name": "Glyph", "args": ()},
    {"class_name": "Guideline", "args": (1,)},
    {"class_name": "Image", "args": ()},
    {"class_name": "ImageSet", "args": ()},
    {"class_name": "Info", "args": ()},
    {"class_name": "Kerning", "args": ()},
    {"class_name": "Layer", "args": ()},
    {
        "class_name": "LayerSet",
        "args": ({"public.default": ufoLib2.objects.Layer()},),
    },
    {"class_name": "Lib", "args": ()},
    {"class_name": "Point", "args": (2, 3)},
]
assert {d["class_name"] for d in BASIC_EMPTY_OBJECTS} == set(ufoLib2.objects.__all__)


@pytest.mark.parametrize("fmt", _SERDE_FORMATS_)
@pytest.mark.parametrize(
    "object_info",
    BASIC_EMPTY_OBJECTS,
    ids=lambda x: x["class_name"],
)
def test_serde_all_objects(fmt: str, object_info: Dict[str, Any]) -> None:
    for req in EXTRAS_REQUIREMENTS[fmt]:
        pytest.importorskip(req)

    klass = getattr(ufoLib2.objects, object_info["class_name"])
    loads = getattr(klass, f"{fmt}_loads")
    obj = klass(*object_info["args"])
    dumps = getattr(obj, f"{fmt}_dumps")
    obj2 = loads(dumps())
    assert obj == obj2
